import json
from ..base import Js
indent = ''
# python 3 support
import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str


def parse(text):
    reviver = arguments[1]
    s = text.to_string().value
    try:
        unfiltered = json.loads(s)
    except:
        raise this.MakeError('SyntaxError',
                             'Could not parse JSON string - Invalid syntax')
    unfiltered = to_js(this, unfiltered)
    if reviver.is_callable():
        root = this.Js({'': unfiltered})
        return walk(root, '', reviver)
    else:
        return unfiltered


def stringify(value, replacer, space):
    global indent
    stack = set([])
    indent = ''
    property_list = replacer_function = this.undefined
    if replacer.is_object():
        if replacer.is_callable():
            replacer_function = replacer
        elif replacer.Class == 'Array':
            property_list = []
            for e in replacer:
                v = replacer[e]
                item = this.undefined
                if v._type() == 'Number':
                    item = v.to_string()
                elif v._type() == 'String':
                    item = v
                elif v.is_object():
                    if v.Class in ('String', 'Number'):
                        item = v.to_string()
                if not item.is_undefined() and item.value not in property_list:
                    property_list.append(item.value)
    if space.is_object():
        if space.Class == 'Number':
            space = space.to_number()
        elif space.Class == 'String':
            space = space.to_string()
    if space._type() == 'Number':
        space = this.Js(min(10, space.to_int()))
        gap = max(int(space.value), 0) * ' '
    elif space._type() == 'String':
        gap = space.value[:10]
    else:
        gap = ''
    return this.Js(
        Str('', this.Js({
            '': value
        }), replacer_function, property_list, gap, stack, space))


def Str(key, holder, replacer_function, property_list, gap, stack, space):
    value = holder[key]
    if value.is_object():
        to_json = value.get('toJSON')
        if to_json.is_callable():
            value = to_json.call(value, (key, ))
    if not replacer_function.is_undefined():
        value = replacer_function.call(holder, (key, value))

    if value.is_object():
        if value.Class == 'String':
            value = value.to_string()
        elif value.Class == 'Number':
            value = value.to_number()
        elif value.Class == 'Boolean':
            value = value.to_boolean()
    if value.is_null():
        return 'null'
    elif value.Class == 'Boolean':
        return 'true' if value.value else 'false'
    elif value._type() == 'String':
        return Quote(value)
    elif value._type() == 'Number':
        if not value.is_infinity():
            return value.to_string()
        return 'null'
    if value.is_object() and not value.is_callable():
        if value.Class == 'Array':
            return ja(value, stack, gap, property_list, replacer_function,
                      space)
        else:
            return jo(value, stack, gap, property_list, replacer_function,
                      space)
    return None  # undefined


def jo(value, stack, gap, property_list, replacer_function, space):
    global indent
    if value in stack:
        raise value.MakeError('TypeError',
                              'Converting circular structure to JSON')
    stack.add(value)
    stepback = indent
    indent += gap
    if not property_list.is_undefined():
        k = property_list
    else:
        k = [e.value for e in value]
    partial = []
    for p in k:
        str_p = value.Js(
            Str(p, value, replacer_function, property_list, gap, stack, space))
        if not str_p.is_undefined():
            member = json.dumps(p) + ':' + (
                ' ' if gap else
                '') + str_p.value  # todo not sure here - what space character?
            partial.append(member)
    if not partial:
        final = '{}'
    else:
        if not gap:
            final = '{%s}' % ','.join(partial)
        else:
            sep = ',\n' + indent
            properties = sep.join(partial)
            final = '{\n' + indent + properties + '\n' + stepback + '}'
    stack.remove(value)
    indent = stepback
    return final


def ja(value, stack, gap, property_list, replacer_function, space):
    global indent
    if value in stack:
        raise value.MakeError('TypeError',
                              'Converting circular structure to JSON')
    stack.add(value)
    stepback = indent
    indent += gap
    partial = []
    length = len(value)
    for index in xrange(length):
        index = str(index)
        str_index = value.Js(
            Str(index, value, replacer_function, property_list, gap, stack,
                space))
        if str_index.is_undefined():
            partial.append('null')
        else:
            partial.append(str_index.value)
    if not partial:
        final = '[]'
    else:
        if not gap:
            final = '[%s]' % ','.join(partial)
        else:
            sep = ',\n' + indent
            properties = sep.join(partial)
            final = '[\n' + indent + properties + '\n' + stepback + ']'
    stack.remove(value)
    indent = stepback
    return final


def Quote(string):
    return string.Js(json.dumps(string.value))


def to_js(this, d):
    if isinstance(d, dict):
        return this.Js(dict((k, this.Js(v)) for k, v in six.iteritems(d)))
    return this.Js(d)


def walk(holder, name, reviver):
    val = holder.get(name)
    if val.Class == 'Array':
        for i in xrange(len(val)):
            i = unicode(i)
            new_element = walk(val, i, reviver)
            if new_element.is_undefined():
                val.delete(i)
            else:
                new_element.put(i, new_element)
    elif val.is_object():
        for key in val:
            new_element = walk(val, key, reviver)
            if new_element.is_undefined():
                val.delete(key)
            else:
                val.put(key, new_element)
    return reviver.call(holder, (name, val))


JSON = Js({})

JSON.define_own_property(
    'parse', {
        'value': Js(parse),
        'enumerable': False,
        'writable': True,
        'configurable': True
    })

JSON.define_own_property(
    'stringify', {
        'value': Js(stringify),
        'enumerable': False,
        'writable': True,
        'configurable': True
    })



import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str

RADIX_SYMBOLS = {
    0: '0',
    1: '1',
    2: '2',
    3: '3',
    4: '4',
    5: '5',
    6: '6',
    7: '7',
    8: '8',
    9: '9',
    10: 'a',
    11: 'b',
    12: 'c',
    13: 'd',
    14: 'e',
    15: 'f',
    16: 'g',
    17: 'h',
    18: 'i',
    19: 'j',
    20: 'k',
    21: 'l',
    22: 'm',
    23: 'n',
    24: 'o',
    25: 'p',
    26: 'q',
    27: 'r',
    28: 's',
    29: 't',
    30: 'u',
    31: 'v',
    32: 'w',
    33: 'x',
    34: 'y',
    35: 'z'
}


def to_str_rep(num):
    if num.is_nan():
        return num.Js('NaN')
    elif num.is_infinity():
        sign = '-' if num.value < 0 else ''
        return num.Js(sign + 'Infinity')
    elif isinstance(num.value,
                    (long, int)) or num.value.is_integer():  # dont print .0
        return num.Js(unicode(int(num.value)))
    return num.Js(unicode(num.value))  # accurate enough


class NumberPrototype:
    def toString(radix):
        if this.Class != 'Number':
            raise this.MakeError('TypeError',
                                 'Number.prototype.valueOf is not generic')
        if radix.is_undefined():
            return to_str_rep(this)
        r = radix.to_int()
        if r == 10:
            return to_str_rep(this)
        if r not in xrange(2, 37):
            raise this.MakeError(
                'RangeError',
                'Number.prototype.toString() radix argument must be between 2 and 36'
            )
        num = this.to_int()
        if num < 0:
            num = -num
            sign = '-'
        else:
            sign = ''
        res = ''
        while num:
            s = RADIX_SYMBOLS[num % r]
            num = num // r
            res = s + res
        return sign + (res if res else '0')

    def valueOf():
        if this.Class != 'Number':
            raise this.MakeError('TypeError',
                                 'Number.prototype.valueOf is not generic')
        return this.value

    def toLocaleString():
        return this.to_string()

    def toFixed(fractionDigits):
        if this.Class != 'Number':
            raise this.MakeError(
                'TypeError',
                'Number.prototype.toFixed called on incompatible receiver')
        digs = fractionDigits.to_int()
        if digs < 0 or digs > 20:
            raise this.MakeError(
                'RangeError',
                'toFixed() digits argument must be between 0 and 20')
        elif this.is_infinity():
            return 'Infinity' if this.value > 0 else '-Infinity'
        elif this.is_nan():
            return 'NaN'
        return format(this.value, '-.%df' % digs)

    def toExponential(fractionDigits):
        if this.Class != 'Number':
            raise this.MakeError(
                'TypeError',
                'Number.prototype.toExponential called on incompatible receiver'
            )
        digs = fractionDigits.to_int()
        if digs < 0 or digs > 20:
            raise this.MakeError(
                'RangeError',
                'toFixed() digits argument must be between 0 and 20')
        elif this.is_infinity():
            return 'Infinity' if this.value > 0 else '-Infinity'
        elif this.is_nan():
            return 'NaN'
        return format(this.value, '-.%de' % digs)

    def toPrecision(precision):
        if this.Class != 'Number':
            raise this.MakeError(
                'TypeError',
                'Number.prototype.toPrecision called on incompatible receiver')
        if precision.is_undefined():
            return this.to_string()
        prec = precision.to_int()
        if this.is_nan():
            return 'NaN'
        elif this.is_infinity():
            return 'Infinity' if this.value > 0 else '-Infinity'
        digs = prec - len(str(int(this.value)))
        if digs >= 0:
            return format(this.value, '-.%df' % digs)
        else:
            return format(this.value, '-.%df' % (prec - 1))



from __future__ import unicode_literals
import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str

#Undefined
class PyJsUndefined(object):
    TYPE = 'Undefined'
    Class = 'Undefined'


undefined = PyJsUndefined()


#Null
class PyJsNull(object):
    TYPE = 'Null'
    Class = 'Null'


null = PyJsNull()

Infinity = float('inf')
NaN = float('nan')

UNDEFINED_TYPE = PyJsUndefined
NULL_TYPE = PyJsNull
STRING_TYPE = unicode if six.PY2 else str
NUMBER_TYPE = float
BOOLEAN_TYPE = bool

# exactly 5 simplexes!
PRIMITIVES = frozenset(
    [UNDEFINED_TYPE, NULL_TYPE, STRING_TYPE, NUMBER_TYPE, BOOLEAN_TYPE])

TYPE_NAMES = {
    UNDEFINED_TYPE: 'Undefined',
    NULL_TYPE: 'Null',
    STRING_TYPE: 'String',
    NUMBER_TYPE: 'Number',
    BOOLEAN_TYPE: 'Boolean',
}


def Type(x):
    # Any -> Str
    return TYPE_NAMES.get(type(x), 'Object')


def GetClass(x):
    # Any -> Str
    cand = TYPE_NAMES.get(type(x))
    if cand is None:
        return x.Class
    return cand


def is_undefined(self):
    return self is undefined


def is_null(self):
    return self is null


def is_primitive(self):
    return type(self) in PRIMITIVES


def is_object(self):
    return not is_primitive(self)


def is_callable(self):
    return hasattr(self, 'call')


def is_infinity(self):
    return self == Infinity or self == -Infinity


def is_nan(self):
    return self != self  # nan!=nan evaluates to True


def is_finite(self):
    return not (is_nan(self) or is_infinity(self))


class JsException(Exception):
    def __init__(self, typ=None, message=None, throw=None):
        if typ is None and message is None and throw is None:
            # it means its the trasnlator based error (old format), do nothing
            self._translator_based = True
        else:
            assert throw is None or (typ is None
                                     and message is None), (throw, typ,
                                                            message)
            self._translator_based = False
            self.typ = typ
            self.message = message
            self.throw = throw

    def get_thrown_value(self, space):
        if self.throw is not None:
            return self.throw
        else:
            return space.NewError(self.typ, self.message)

    def __str__(self):
        if self._translator_based:
            if self.mes.Class == 'Error':
                return self.mes.callprop('toString').value
            else:
                return self.mes.to_string().value
        else:
            if self.throw is not None:
                from .conversions import to_string
                return to_string(self.throw)
            else:
                return self.typ + ': ' + self.message


def MakeError(typ, message=u'no info', throw=None):
    return JsException(typ,
                       unicode(message) if message is not None else message,
                       throw)


def value_from_js_exception(js_exception, space):
    if js_exception.throw is not None:
        return js_exception.throw
    else:
        return space.NewError(js_exception.typ, js_exception.message)


def js_dtoa(number):
    if is_nan(number):
        return u'NaN'
    elif is_infinity(number):
        if number > 0:
            return u'Infinity'
        return u'-Infinity'
    elif number == 0.:
        return u'0'
    elif abs(number) < 1e-6 or abs(number) >= 1e21:
        frac, exponent = unicode(repr(float(number))).split('e')
        # Remove leading zeros from the exponent.
        exponent = int(exponent)
        return frac + ('e' if exponent < 0 else 'e+') + unicode(exponent)
    elif abs(number) < 1e-4:  # python starts to return exp notation while we still want the prec
        frac, exponent = unicode(repr(float(number))).split('e-')
        base = u'0.' + u'0' * (int(exponent) - 1) + frac.lstrip('-').replace('.', '')
        return base if number > 0. else u'-' + base
    elif isinstance(number, long) or number.is_integer():  # dont print .0
        return unicode(int(number))
    return unicode(repr(number))  # python representation should be equivalent.



from __future__ import unicode_literals

from ..conversions import *
from ..func_utils import *


def Number(this, args):
    if len(args) == 0:
        return 0.
    return to_number(args[0])


def NumberConstructor(args, space):
    temp = space.NewObject()
    temp.prototype = space.NumberPrototype
    temp.Class = 'Number'
    temp.value = float(to_number(get_arg(args, 0)) if len(args) > 0 else 0.)
    return temp


CONSTS = {
    'MAX_VALUE': 1.7976931348623157e308,
    'MIN_VALUE': 5.0e-324,
    'NaN': NaN,
    'NEGATIVE_INFINITY': Infinity,
    'POSITIVE_INFINITY': -Infinity
}



from __future__ import unicode_literals
from ..conversions import *
from ..func_utils import *


class ObjectPrototype:
    def toString(this, args):
        if type(this) == UNDEFINED_TYPE:
            return u'[object Undefined]'
        elif type(this) == NULL_TYPE:
            return u'[object Null]'
        return u'[object %s]' % GetClass(to_object(this, args.space))

    def valueOf(this, args):
        return to_object(this, args.space)

    def toLocaleString(this, args):
        o = to_object(this, args.space)
        toString = o.get(u'toString')
        if not is_callable(toString):
            raise MakeError('TypeError', 'toString of this is not callcable')
        else:
            return toString.call(this, args)

    def hasOwnProperty(this, args):
        prop = get_arg(args, 0)
        o = to_object(this, args.space)
        return o.get_own_property(to_string(prop)) is not None

    def isPrototypeOf(this, args):
        # a bit stupid specification because of object conversion that will cause invalid values for primitives
        # for example Object.prototype.isPrototypeOf.call((5).__proto__, 5) gives false
        obj = get_arg(args, 0)
        if not is_object(obj):
            return False
        o = to_object(this, args.space)
        while 1:
            obj = obj.prototype
            if obj is None or is_null(obj):
                return False
            if obj is o:
                return True

    def propertyIsEnumerable(this, args):
        prop = get_arg(args, 0)
        o = to_object(this, args.space)
        cand = o.own.get(to_string(prop))
        return cand is not None and cand.get('enumerable')



class ObjectPrototype:
    def toString():
        return '[object %s]' % this.Class

    def valueOf():
        return this.to_object()

    def toLocaleString():
        return this.callprop('toString')

    def hasOwnProperty(prop):
        return this.get_own_property(prop.to_string().value) is not None

    def isPrototypeOf(obj):
        #a bit stupid specification but well
        # for example Object.prototype.isPrototypeOf.call((5).__proto__, 5) gives false
        if not obj.is_object():
            return False
        while 1:
            obj = obj.prototype
            if obj is None or obj.is_null():
                return False
            if obj is this:
                return True

    def propertyIsEnumerable(prop):
        cand = this.own.get(prop.to_string().value)
        return cand is not None and cand.get('enumerable')



class ErrorPrototype:
    def toString():
        if this.TYPE != 'Object':
            raise this.MakeError(
                'TypeError', 'Error.prototype.toString called on non-object')
        name = this.get('name')
        name = 'Error' if name.is_undefined() else name.to_string().value
        msg = this.get('message')
        msg = '' if msg.is_undefined() else msg.to_string().value
        return name + (name and msg and ': ') + msg



from ..base import *
import six

#todo Double check everything is OK


@Js
def Object():
    val = arguments.get('0')
    if val.is_null() or val.is_undefined():
        return PyJsObject(prototype=ObjectPrototype)
    return val.to_object()


@Js
def object_constructor():
    if len(arguments):
        val = arguments.get('0')
        if val.TYPE == 'Object':
            #Implementation dependent, but my will simply return :)
            return val
        elif val.TYPE in ('Number', 'String', 'Boolean'):
            return val.to_object()
    return PyJsObject(prototype=ObjectPrototype)


Object.create = object_constructor
Object.own['length']['value'] = Js(1)


class ObjectMethods:
    def getPrototypeOf(obj):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.getPrototypeOf called on non-object')
        return null if obj.prototype is None else obj.prototype

    def getOwnPropertyDescriptor(obj, prop):
        if not obj.is_object():
            raise MakeError(
                'TypeError',
                'Object.getOwnPropertyDescriptor called on non-object')
        return obj.own.get(
            prop.to_string().
            value)  # will return undefined if we dont have this prop

    def getOwnPropertyNames(obj):
        if not obj.is_object():
            raise MakeError(
                'TypeError',
                'Object.getOwnPropertyDescriptor called on non-object')
        return obj.own.keys()

    def create(obj):
        if not (obj.is_object() or obj.is_null()):
            raise MakeError('TypeError',
                            'Object prototype may only be an Object or null')
        temp = PyJsObject(prototype=(None if obj.is_null() else obj))
        if len(arguments) > 1 and not arguments[1].is_undefined():
            if six.PY2:
                ObjectMethods.defineProperties.__func__(temp, arguments[1])
            else:
                ObjectMethods.defineProperties(temp, arguments[1])
        return temp

    def defineProperty(obj, prop, attrs):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.defineProperty called on non-object')
        name = prop.to_string().value
        if not obj.define_own_property(name, ToPropertyDescriptor(attrs)):
            raise MakeError('TypeError', 'Cannot redefine property: %s' % name)
        return obj

    def defineProperties(obj, properties):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.defineProperties called on non-object')
        props = properties.to_object()
        for name in props:
            desc = ToPropertyDescriptor(props.get(name.value))
            if not obj.define_own_property(name.value, desc):
                raise MakeError(
                    'TypeError',
                    'Failed to define own property: %s' % name.value)
        return obj

    def seal(obj):
        if not obj.is_object():
            raise MakeError('TypeError', 'Object.seal called on non-object')
        for desc in obj.own.values():
            desc['configurable'] = False
        obj.extensible = False
        return obj

    def freeze(obj):
        if not obj.is_object():
            raise MakeError('TypeError', 'Object.freeze called on non-object')
        for desc in obj.own.values():
            desc['configurable'] = False
            if is_data_descriptor(desc):
                desc['writable'] = False
        obj.extensible = False
        return obj

    def preventExtensions(obj):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.preventExtensions on non-object')
        obj.extensible = False
        return obj

    def isSealed(obj):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.isSealed called on non-object')
        if obj.extensible:
            return False
        for desc in obj.own.values():
            if desc['configurable']:
                return False
        return True

    def isFrozen(obj):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.isFrozen called on non-object')
        if obj.extensible:
            return False
        for desc in obj.own.values():
            if desc['configurable']:
                return False
            if is_data_descriptor(desc) and desc['writable']:
                return False
        return True

    def isExtensible(obj):
        if not obj.is_object():
            raise MakeError('TypeError',
                            'Object.isExtensible called on non-object')
        return obj.extensible

    def keys(obj):
        if not obj.is_object():
            raise MakeError('TypeError', 'Object.keys called on non-object')
        return [e for e, d in six.iteritems(obj.own) if d.get('enumerable')]


# add methods attached to Object constructor
fill_prototype(Object, ObjectMethods, default_attrs)
# add constructor to prototype
fill_in_props(ObjectPrototype, {'constructor': Object}, default_attrs)
# add prototype property to the constructor.
Object.define_own_property(
    'prototype', {
        'value': ObjectPrototype,
        'enumerable': False,
        'writable': False,
        'configurable': False
    })

# some utility functions:


def ToPropertyDescriptor(obj):  # page 38 (50 absolute)
    if obj.TYPE != 'Object':
        raise MakeError('TypeError',
                        'Can\'t convert non-object to property descriptor')
    desc = {}
    if obj.has_property('enumerable'):
        desc['enumerable'] = obj.get('enumerable').to_boolean().value
    if obj.has_property('configurable'):
        desc['configurable'] = obj.get('configurable').to_boolean().value
    if obj.has_property('value'):
        desc['value'] = obj.get('value')
    if obj.has_property('writable'):
        desc['writable'] = obj.get('writable').to_boolean().value
    if obj.has_property('get'):
        cand = obj.get('get')
        if not (cand.is_undefined() or cand.is_callable()):
            raise MakeError(
                'TypeError',
                'Invalid getter (it has to be a function or undefined)')
        desc['get'] = cand
    if obj.has_property('set'):
        cand = obj.get('set')
        if not (cand.is_undefined() or cand.is_callable()):
            raise MakeError(
                'TypeError',
                'Invalid setter (it has to be a function or undefined)')
        desc['set'] = cand
    if ('get' in desc or 'set' in desc) and ('value' in desc
                                             or 'writable' in desc):
        raise MakeError(
            'TypeError',
            'Invalid property.  A property cannot both have accessors and be writable or have a value.'
        )
    return desc



from __future__ import unicode_literals

from .base import Scope
from .func_utils import *
from .conversions import *
import six
from .prototypes.jsboolean import BooleanPrototype
from .prototypes.jserror import ErrorPrototype
from .prototypes.jsfunction import FunctionPrototype
from .prototypes.jsnumber import NumberPrototype
from .prototypes.jsobject import ObjectPrototype
from .prototypes.jsregexp import RegExpPrototype
from .prototypes.jsstring import StringPrototype
from .prototypes.jsarray import ArrayPrototype
from .prototypes import jsjson
from .prototypes import jsutils

from .constructors import jsnumber, jsstring, jsarray, jsboolean, jsregexp, jsmath, jsobject, jsfunction, jsconsole



def fill_proto(proto, proto_class, space):
    for i in dir(proto_class):
        e = getattr(proto_class, i)
        if six.PY2:
            if hasattr(e, '__func__'):
                meth = e.__func__
            else:
                continue
        else:
            if hasattr(e, '__call__') and not i.startswith('__'):
                meth = e
            else:
                continue
        meth_name = meth.__name__.strip('_')  # RexExp._exec -> RegExp.exec
        js_meth = space.NewFunction(meth, space.ctx, (), meth_name, False, ())
        set_non_enumerable(proto, meth_name, js_meth)
    return proto


def easy_func(f, space):
    return space.NewFunction(f, space.ctx, (), f.__name__, False, ())


def Empty(this, args):
    return undefined


def set_non_enumerable(obj, name, prop):
    obj.define_own_property(
        unicode(name), {
            'value': prop,
            'writable': True,
            'enumerable': False,
            'configurable': True
        }, True)


def set_protected(obj, name, prop):
    obj.define_own_property(
        unicode(name), {
            'value': prop,
            'writable': False,
            'enumerable': False,
            'configurable': False
        }, True)


def fill_space(space, byte_generator):
    # set global scope
    global_scope = Scope({}, space, parent=None)
    global_scope.THIS_BINDING = global_scope
    global_scope.registers(byte_generator.declared_vars)
    space.GlobalObj = global_scope

    space.byte_generator = byte_generator

    # first init all protos, later take care of constructors and details

    # Function must be first obviously, we have to use a small trick to do that...
    function_proto = space.NewFunction(Empty, space.ctx, (), 'Empty', False,
                                       ())
    space.FunctionPrototype = function_proto  # this will fill the prototypes of the methods!
    fill_proto(function_proto, FunctionPrototype, space)

    # Object next
    object_proto = space.NewObject()  # no proto
    fill_proto(object_proto, ObjectPrototype, space)
    space.ObjectPrototype = object_proto
    function_proto.prototype = object_proto

    # Number
    number_proto = space.NewObject()
    number_proto.prototype = object_proto
    fill_proto(number_proto, NumberPrototype, space)
    number_proto.value = 0.
    number_proto.Class = 'Number'
    space.NumberPrototype = number_proto

    # String
    string_proto = space.NewObject()
    string_proto.prototype = object_proto
    fill_proto(string_proto, StringPrototype, space)
    string_proto.value = u''
    string_proto.Class = 'String'
    space.StringPrototype = string_proto

    # Boolean
    boolean_proto = space.NewObject()
    boolean_proto.prototype = object_proto
    fill_proto(boolean_proto, BooleanPrototype, space)
    boolean_proto.value = False
    boolean_proto.Class = 'Boolean'
    space.BooleanPrototype = boolean_proto

    # Array
    array_proto = space.NewArray(0)
    array_proto.prototype = object_proto
    fill_proto(array_proto, ArrayPrototype, space)
    space.ArrayPrototype = array_proto

    # JSON
    json = space.NewObject()
    json.put(u'stringify', easy_func(jsjson.stringify, space))
    json.put(u'parse', easy_func(jsjson.parse, space))

    # Utils
    parseFloat = easy_func(jsutils.parseFloat, space)
    parseInt = easy_func(jsutils.parseInt, space)
    isNaN = easy_func(jsutils.isNaN, space)
    isFinite = easy_func(jsutils.isFinite, space)

    # Error
    error_proto = space.NewError(u'Error', u'')
    error_proto.prototype = object_proto
    error_proto.put(u'name', u'Error')
    fill_proto(error_proto, ErrorPrototype, space)
    space.ErrorPrototype = error_proto

    def construct_constructor(typ):
        def creator(this, args):
            message = get_arg(args, 0)
            if not is_undefined(message):
                msg = to_string(message)
            else:
                msg = u''
            return space.NewError(typ, msg)

        j = easy_func(creator, space)
        j.name = unicode(typ)

        set_protected(j, 'prototype', space.ERROR_TYPES[typ])

        set_non_enumerable(space.ERROR_TYPES[typ], 'constructor', j)

        def new_create(args, space):
            message = get_arg(args, 0)
            if not is_undefined(message):
                msg = to_string(message)
            else:
                msg = u''
            return space.NewError(typ, msg)

        j.create = new_create
        return j

    # fill remaining error types
    error_constructors = {}
    for err_type_name in (u'Error', u'EvalError', u'RangeError',
                          u'ReferenceError', u'SyntaxError', u'TypeError',
                          u'URIError'):
        extra_err = space.NewError(u'Error', u'')
        extra_err.put(u'name', err_type_name)
        setattr(space, err_type_name + u'Prototype', extra_err)
        error_constructors[err_type_name] = construct_constructor(
            err_type_name)

    assert space.TypeErrorPrototype is not None

    # RegExp
    regexp_proto = space.NewRegExp(u'(?:)', u'')
    regexp_proto.prototype = object_proto
    fill_proto(regexp_proto, RegExpPrototype, space)
    space.RegExpPrototype = regexp_proto

    # Json

    # now all these boring constructors...

    # Number
    number = easy_func(jsnumber.Number, space)
    space.Number = number
    number.create = jsnumber.NumberConstructor
    set_non_enumerable(number_proto, 'constructor', number)
    set_protected(number, 'prototype', number_proto)
    # number has some extra constants
    for k, v in jsnumber.CONSTS.items():
        set_protected(number, k, v)

    # String
    string = easy_func(jsstring.String, space)
    space.String = string
    string.create = jsstring.StringConstructor
    set_non_enumerable(string_proto, 'constructor', string)
    set_protected(string, 'prototype', string_proto)
    # string has an extra function
    set_non_enumerable(string, 'fromCharCode',
                       easy_func(jsstring.fromCharCode, space))

    # Boolean
    boolean = easy_func(jsboolean.Boolean, space)
    space.Boolean = boolean
    boolean.create = jsboolean.BooleanConstructor
    set_non_enumerable(boolean_proto, 'constructor', boolean)
    set_protected(boolean, 'prototype', boolean_proto)

    # Array
    array = easy_func(jsarray.Array, space)
    space.Array = array
    array.create = jsarray.ArrayConstructor
    set_non_enumerable(array_proto, 'constructor', array)
    set_protected(array, 'prototype', array_proto)
    array.put(u'isArray', easy_func(jsarray.isArray, space))

    # RegExp
    regexp = easy_func(jsregexp.RegExp, space)
    space.RegExp = regexp
    regexp.create = jsregexp.RegExpCreate
    set_non_enumerable(regexp_proto, 'constructor', regexp)
    set_protected(regexp, 'prototype', regexp_proto)

    # Object
    _object = easy_func(jsobject.Object, space)
    space.Object = _object
    _object.create = jsobject.ObjectCreate
    set_non_enumerable(object_proto, 'constructor', _object)
    set_protected(_object, 'prototype', object_proto)
    fill_proto(_object, jsobject.ObjectMethods, space)

    # Function
    function = easy_func(jsfunction.Function, space)
    space.Function = function

    # Math
    math = space.NewObject()
    math.Class = 'Math'
    fill_proto(math, jsmath.MathFunctions, space)
    for k, v in jsmath.CONSTANTS.items():
        set_protected(math, k, v)

    console = space.NewObject()
    fill_proto(console, jsconsole.ConsoleMethods, space)

    # set global object
    builtins = {
        'String': string,
        'Number': number,
        'Boolean': boolean,
        'RegExp': regexp,
        'exports': convert_to_js_type({}, space),
        'Math': math,
        #'Date',
        'Object': _object,
        'Function': function,
        'JSON': json,
        'Array': array,
        'parseFloat': parseFloat,
        'parseInt': parseInt,
        'isFinite': isFinite,
        'isNaN': isNaN,
        'eval': easy_func(jsfunction._eval, space),
        'console': console,
        'log': console.get(u'log'),
    }

    builtins.update(error_constructors)

    set_protected(global_scope, 'NaN', NaN)
    set_protected(global_scope, 'Infinity', Infinity)
    for k, v in builtins.items():
        set_non_enumerable(global_scope, k, v)



# this is based on jsarray.py

from ..base import *
try:
    import numpy
except:
    pass


@Js
def Int16Array():
    TypedArray = (PyJsInt8Array, PyJsUint8Array, PyJsUint8ClampedArray,
                  PyJsInt16Array, PyJsUint16Array, PyJsInt32Array,
                  PyJsUint32Array, PyJsFloat32Array, PyJsFloat64Array)
    a = arguments[0]
    if isinstance(a, PyJsNumber):  # length
        length = a.to_uint32()
        if length != a.value:
            raise MakeError('RangeError', 'Invalid array length')
        temp = Js(numpy.full(length, 0, dtype=numpy.int16))
        temp.put('length', a)
        return temp
    elif isinstance(a, PyJsString):  # object (string)
        temp = Js(numpy.array(list(a.value), dtype=numpy.int16))
        temp.put('length', Js(len(list(a.value))))
        return temp
    elif isinstance(a, PyJsArray) or isinstance(a, TypedArray) or isinstance(
            a, PyJsArrayBuffer):  # object (Array, TypedArray)
        array = a.to_list()
        array = [(int(item.value) if item.value != None else 0)
                 for item in array]
        temp = Js(numpy.array(array, dtype=numpy.int16))
        temp.put('length', Js(len(array)))
        return temp
    elif isinstance(a, PyObjectWrapper):  # object (ArrayBuffer, etc)
        if len(a.obj) % 2 != 0:
            raise MakeError(
                'RangeError',
                'Byte length of Int16Array should be a multiple of 2')
        if len(arguments) > 1:
            offset = int(arguments[1].value)
            if offset % 2 != 0:
                raise MakeError(
                    'RangeError',
                    'Start offset of Int16Array should be a multiple of 2')
        else:
            offset = 0
        if len(arguments) > 2:
            length = int(arguments[2].value)
        else:
            length = int((len(a.obj) - offset) / 2)
        array = numpy.frombuffer(
            a.obj, dtype=numpy.int16, count=length, offset=offset)
        temp = Js(array)
        temp.put('length', Js(length))
        temp.buff = array
        return temp
    temp = Js(numpy.full(0, 0, dtype=numpy.int16))
    temp.put('length', Js(0))
    return temp


Int16Array.create = Int16Array
Int16Array.own['length']['value'] = Js(3)

Int16Array.define_own_property(
    'prototype', {
        'value': Int16ArrayPrototype,
        'enumerable': False,
        'writable': False,
        'configurable': False
    })

Int16ArrayPrototype.define_own_property(
    'constructor', {
        'value': Int16Array,
        'enumerable': False,
        'writable': True,
        'configurable': True
    })

Int16ArrayPrototype.define_own_property('BYTES_PER_ELEMENT', {
    'value': Js(2),
    'enumerable': False,
    'writable': False,
    'configurable': False
})



from pyjsparser.pyjsparserdata import *

REGEXP_SPECIAL_SINGLE = {'\\', '^', '$', '*', '+', '?', '.'}

NOT_PATTERN_CHARS = {
    '^', '$', '\\', '.', '*', '+', '?', '(', ')', '[', ']', '|'
}  # what about '{', '}',  ???

CHAR_CLASS_ESCAPE = {'d', 'D', 's', 'S', 'w', 'W'}
CONTROL_ESCAPE_CHARS = {'f', 'n', 'r', 't', 'v'}
CONTROL_LETTERS = {
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z'
}


def SpecialChar(char):
    return {'type': 'SpecialChar', 'content': char}


def isPatternCharacter(char):
    return char not in NOT_PATTERN_CHARS


class JsRegExpParser:
    def __init__(self, source, flags):
        self.source = source
        self.flags = flags
        self.index = 0
        self.length = len(source)
        self.lineNumber = 0
        self.lineStart = 0

    def parsePattern(self):
        '''Perform sctring escape - for regexp literals'''
        return {'type': 'Pattern', 'contents': self.parseDisjunction()}

    def parseDisjunction(self):
        alternatives = []
        while True:
            alternatives.append(self.parseAlternative())
            if not self.isEOF():
                self.expect_character('|')
            else:
                break
        return {'type': 'Disjunction', 'contents': alternatives}

    def isEOF(self):
        if self.index >= self.length:
            return True
        return False

    def expect_character(self, character):
        if self.source[self.index] != character:
            self.throwUnexpected(character)
        self.index += 1

    def parseAlternative(self):
        contents = []
        while not self.isEOF() and self.source[self.index] != '|':
            contents.append(self.parseTerm())
        return {'type': 'Alternative', 'contents': contents}

    def follows(self, chars):
        for i, c in enumerate(chars):
            if self.index + i >= self.length or self.source[self.index +
                                                            i] != c:
                return False
        return True

    def parseTerm(self):
        assertion = self.parseAssertion()
        if assertion:
            return assertion
        else:
            return {
                'type': 'Term',
                'contents': self.parseAtom()
            }  # quantifier will go inside atom!

    def parseAssertion(self):
        if self.follows('$'):
            content = SpecialChar('$')
            self.index += 1
        elif self.follows('^'):
            content = SpecialChar('^')
            self.index += 1
        elif self.follows('\\b'):
            content = SpecialChar('\\b')
            self.index += 2
        elif self.follows('\\B'):
            content = SpecialChar('\\B')
            self.index += 2
        elif self.follows('(?='):
            self.index += 3
            dis = self.parseDisjunction()
            self.expect_character(')')
            content = {'type': 'Lookached', 'contents': dis, 'negated': False}
        elif self.follows('(?!'):
            self.index += 3
            dis = self.parseDisjunction()
            self.expect_character(')')
            content = {'type': 'Lookached', 'contents': dis, 'negated': True}
        else:
            return None
        return {'type': 'Assertion', 'content': content}

    def parseAtom(self):
        if self.follows('.'):
            content = SpecialChar('.')
            self.index += 1
        elif self.follows('\\'):
            self.index += 1
            content = self.parseAtomEscape()
        elif self.follows('['):
            content = self.parseCharacterClass()
        elif self.follows('(?:'):
            self.index += 3
            dis = self.parseDisjunction()
            self.expect_character(')')
            content = 'idk'
        elif self.follows('('):
            self.index += 1
            dis = self.parseDisjunction()
            self.expect_character(')')
            content = 'idk'
        elif isPatternCharacter(self.source[self.index]):
            content = self.source[self.index]
            self.index += 1
        else:
            return None
        quantifier = self.parseQuantifier()
        return {'type': 'Atom', 'content': content, 'quantifier': quantifier}

    def parseQuantifier(self):
        prefix = self.parseQuantifierPrefix()
        if not prefix:
            return None
        greedy = True
        if self.follows('?'):
            self.index += 1
            greedy = False
        return {'type': 'Quantifier', 'contents': prefix, 'greedy': greedy}

    def parseQuantifierPrefix(self):
        if self.isEOF():
            return None
        if self.follows('+'):
            content = '+'
            self.index += 1
        elif self.follows('?'):
            content = '?'
            self.index += 1
        elif self.follows('*'):
            content = '*'
            self.index += 1
        elif self.follows(
                '{'
        ):  # try matching otherwise return None and restore the state
            i = self.index
            self.index += 1
            digs1 = self.scanDecimalDigs()
            # if no minimal number of digs provided then return no quantifier
            if not digs1:
                self.index = i
                return None
            # scan char limit if provided
            if self.follows(','):
                self.index += 1
                digs2 = self.scanDecimalDigs()
            else:
                digs2 = ''
            # must be valid!
            if not self.follows('}'):
                self.index = i
                return None
            else:
                self.expect_character('}')
                content = int(digs1), int(digs2) if digs2 else None
        else:
            return None
        return content

    def parseAtomEscape(self):
        ch = self.source[self.index]
        if isDecimalDigit(ch) and ch != 0:
            digs = self.scanDecimalDigs()
        elif ch in CHAR_CLASS_ESCAPE:
            self.index += 1
            return SpecialChar('\\' + ch)
        else:
            return self.parseCharacterEscape()

    def parseCharacterEscape(self):
        ch = self.source[self.index]
        if ch in CONTROL_ESCAPE_CHARS:
            return SpecialChar('\\' + ch)
        if ch == 'c':
            'ok, fuck this shit.'

    def scanDecimalDigs(self):
        s = self.index
        while not self.isEOF() and isDecimalDigit(self.source[self.index]):
            self.index += 1
        return self.source[s:self.index]


a = JsRegExpParser('a(?=x)', '')
print(a.parsePattern())



from .opcodes import *
from .space import *
from .base import *


class Code:
    '''Can generate, store and run sequence of ops representing js code'''

    def __init__(self, is_strict=False, debug_mode=False):
        self.tape = []
        self.compiled = False
        self.label_locs = None
        self.is_strict = is_strict
        self.debug_mode = debug_mode

        self.contexts = []
        self.current_ctx = None
        self.return_locs = []
        self._label_count = 0
        self.label_locs = None

        # useful references
        self.GLOBAL_THIS = None
        self.space = None

        # dbg
        self.ctx_depth = 0


    def get_new_label(self):
        self._label_count += 1
        return self._label_count

    def emit(self, op_code, *args):
        ''' Adds op_code with specified args to tape '''
        self.tape.append(OP_CODES[op_code](*args))

    def compile(self, start_loc=0):
        ''' Records locations of labels and compiles the code '''
        self.label_locs = {} if self.label_locs is None else self.label_locs
        loc = start_loc
        while loc < len(self.tape):
            if type(self.tape[loc]) == LABEL:
                self.label_locs[self.tape[loc].num] = loc
                del self.tape[loc]
                continue
            loc += 1
        self.compiled = True

    def _call(self, func, this, args):
        ''' Calls a bytecode function func
            NOTE:  use !ONLY! when calling functions from native methods! '''
        assert not func.is_native
        # fake call - the the runner to return to the end of the file
        old_contexts = self.contexts
        old_return_locs = self.return_locs
        old_curr_ctx = self.current_ctx

        self.contexts = [FakeCtx()]
        self.return_locs = [len(self.tape)]  # target line after return

        # prepare my ctx
        my_ctx = func._generate_my_context(this, args)
        self.current_ctx = my_ctx

        # execute dunction
        ret = self.run(my_ctx, starting_loc=self.label_locs[func.code])

        # bring back old execution
        self.current_ctx = old_curr_ctx
        self.contexts = old_contexts
        self.return_locs = old_return_locs

        return ret

    def execute_fragment_under_context(self, ctx, start_label, end_label):
        ''' just like run but returns if moved outside of the specified fragment
            # 4 different exectution results
            # 0=normal, 1=return, 2=jump_outside, 3=errors
            # execute_fragment_under_context returns:
            # (return_value, typ, return_value/jump_loc/py_error)
            # IMPARTANT: It is guaranteed that the length of the ctx.stack is unchanged.
        '''
        old_curr_ctx = self.current_ctx
        self.ctx_depth += 1
        old_stack_len = len(ctx.stack)
        old_ret_len = len(self.return_locs)
        old_ctx_len = len(self.contexts)
        try:
            self.current_ctx = ctx
            return self._execute_fragment_under_context(
                ctx, start_label, end_label)
        except JsException as err:
            if self.debug_mode:
                self._on_fragment_exit("js errors")
            # undo the things that were put on the stack (if any) to ensure a proper error recovery
            del ctx.stack[old_stack_len:]
            del self.return_locs[old_ret_len:]
            del self.contexts[old_ctx_len :]
            return undefined, 3, err
        finally:
            self.ctx_depth -= 1
            self.current_ctx = old_curr_ctx
            assert old_stack_len == len(ctx.stack)

    def _get_dbg_indent(self):
        return self.ctx_depth * '  '

    def _on_fragment_exit(self, mode):
        print(self._get_dbg_indent() + 'ctx exit (%s)' % mode)

    def _execute_fragment_under_context(self, ctx, start_label, end_label):
        start, end = self.label_locs[start_label], self.label_locs[end_label]
        initial_len = len(ctx.stack)
        loc = start
        entry_level = len(self.contexts)
        # for e in self.tape[start:end]:
        #     print e
        if self.debug_mode:
            print(self._get_dbg_indent() + 'ctx entry (from:%d, to:%d)' % (start, end))
        while loc < len(self.tape):
            if len(self.contexts) == entry_level and loc >= end:
                if self.debug_mode:
                    self._on_fragment_exit('normal')
                assert loc == end
                delta_stack = len(ctx.stack) - initial_len
                assert delta_stack == +1, 'Stack change must be equal to +1! got %d' % delta_stack
                return ctx.stack.pop(), 0, None  # means normal return

            # execute instruction
            if self.debug_mode:
                print(self._get_dbg_indent() + str(loc), self.tape[loc])
            status = self.tape[loc].eval(ctx)

            # check status for special actions
            if status is not None:
                if type(status) == int:  # jump to label
                    loc = self.label_locs[status]
                    if len(self.contexts) == entry_level:
                        # check if jumped outside of the fragment and break if so
                        if not start <= loc < end:
                            if self.debug_mode:
                                self._on_fragment_exit('jump outside loc:%d label:%d' % (loc, status))
                            delta_stack = len(ctx.stack) - initial_len
                            assert delta_stack == +1, 'Stack change must be equal to +1! got %d' % delta_stack
                            return ctx.stack.pop(), 2, status  # jump outside
                    continue

                elif len(status) == 2:  # a call or a return!
                    # call: (new_ctx, func_loc_label_num)
                    if status[0] is not None:
                        # append old state to the stack
                        self.contexts.append(ctx)
                        self.return_locs.append(loc + 1)
                        # set new state
                        loc = self.label_locs[status[1]]
                        ctx = status[0]
                        self.current_ctx = ctx
                        continue

                    # return: (None, None)
                    else:
                        if len(self.contexts) == entry_level:
                            if self.debug_mode:
                                self._on_fragment_exit('return')
                            delta_stack = len(ctx.stack) - initial_len
                            assert delta_stack == +1, 'Stack change must be equal to +1! got %d' % delta_stack
                            return undefined, 1, ctx.stack.pop(
                            )  # return signal
                        return_value = ctx.stack.pop()
                        ctx = self.contexts.pop()
                        self.current_ctx = ctx
                        ctx.stack.append(return_value)

                        loc = self.return_locs.pop()
                        continue
            # next instruction
            loc += 1
        if self.debug_mode:
            self._on_fragment_exit('internal error - unexpected end of tape, will crash')
        assert False, 'Remember to add NOP at the end!'

    def run(self, ctx, starting_loc=0):
        loc = starting_loc
        self.current_ctx = ctx
        while loc < len(self.tape):
            # execute instruction
            if self.debug_mode:
                print(loc, self.tape[loc])
            status = self.tape[loc].eval(ctx)

            # check status for special actions
            if status is not None:
                if type(status) == int:  # jump to label
                    loc = self.label_locs[status]
                    continue

                elif len(status) == 2:  # a call or a return!
                    # call: (new_ctx, func_loc_label_num)
                    if status[0] is not None:
                        # append old state to the stack
                        self.contexts.append(ctx)
                        self.return_locs.append(loc + 1)
                        # set new state
                        loc = self.label_locs[status[1]]
                        ctx = status[0]
                        self.current_ctx = ctx
                        continue

                    # return: (None, None)
                    else:
                        return_value = ctx.stack.pop()
                        ctx = self.contexts.pop()
                        self.current_ctx = ctx
                        ctx.stack.append(return_value)

                        loc = self.return_locs.pop()
                        continue
            # next instruction
            loc += 1
        assert len(ctx.stack) == 1, ctx.stack
        return ctx.stack.pop()


class FakeCtx(object):
    def __init__(self):
        self.stack = []



from __future__ import unicode_literals

from ..conversions import *
from ..func_utils import *

# python 3 support
import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str

# todo fix apply and bind


class FunctionPrototype:
    def toString(this, args):
        if not is_callable(this):
            raise MakeError('TypeError',
                            'Function.prototype.toString is not generic')

        args = u', '.join(map(unicode, this.params))
        return u'function %s(%s) { [native code] }' % (this.name if this.name
                                                       else u'', args)

    def call(this, args):
        if not is_callable(this):
            raise MakeError('TypeError',
                            'Function.prototype.call is not generic')
        _this = get_arg(args, 0)
        _args = tuple(args)[1:]
        return this.call(_this, _args)

    def apply(this, args):
        if not is_callable(this):
            raise MakeError('TypeError',
                            'Function.prototype.apply is not generic')
        _args = get_arg(args, 1)
        if not is_object(_args):
            raise MakeError(
                'TypeError',
                'argList argument to Function.prototype.apply must an Object')
        _this = get_arg(args, 0)
        return this.call(_this, js_array_to_tuple(_args))

    def bind(this, args):
        if not is_callable(this):
            raise MakeError('TypeError',
                            'Function.prototype.bind is not generic')
        bound_this = get_arg(args, 0)
        bound_args = tuple(args)[1:]

        def bound(dummy_this, extra_args):
            return this.call(bound_this, bound_args + tuple(extra_args))

        js_bound = args.space.NewFunction(bound, this.ctx, (), u'', False, ())
        js_bound.put(u'length',
                     float(max(len(this.params) - len(bound_args), 0.)))
        js_bound.put(u'name', u'boundFunc')
        return js_bound



from ..conversions import *
from ..func_utils import *
from six import unichr

def fromCharCode(this, args):
    res = u''
    for e in args:
        res += unichr(to_uint16(e))
    return res


def String(this, args):
    if len(args) == 0:
        return u''
    return to_string(args[0])


def StringConstructor(args, space):
    temp = space.NewObject()
    temp.prototype = space.StringPrototype
    temp.Class = 'String'
    temp.value = to_string(get_arg(args, 0)) if len(args) > 0 else u''
    return temp



from .base import *
from .constructors.jsmath import Math
from .constructors.jsdate import Date
from .constructors.jsobject import Object
from .constructors.jsfunction import Function
from .constructors.jsstring import String
from .constructors.jsnumber import Number
from .constructors.jsboolean import Boolean
from .constructors.jsregexp import RegExp
from .constructors.jsarray import Array
from .constructors.jsarraybuffer import ArrayBuffer
from .constructors.jsint8array import Int8Array
from .constructors.jsuint8array import Uint8Array
from .constructors.jsuint8clampedarray import Uint8ClampedArray
from .constructors.jsint16array import Int16Array
from .constructors.jsuint16array import Uint16Array
from .constructors.jsint32array import Int32Array
from .constructors.jsuint32array import Uint32Array
from .constructors.jsfloat32array import Float32Array
from .constructors.jsfloat64array import Float64Array
from .prototypes.jsjson import JSON
from .host.console import console
from .host.jseval import Eval
from .host.jsfunctions import parseFloat, parseInt, isFinite, \
    isNaN, escape, unescape, encodeURI, decodeURI, encodeURIComponent, decodeURIComponent

# Now we have all the necessary items to create global environment for script
__all__ = [
    'Js', 'PyJsComma', 'PyJsStrictEq', 'PyJsStrictNeq', 'PyJsException',
    'PyJsBshift', 'Scope', 'PyExceptionToJs', 'JsToPyException', 'JS_BUILTINS',
    'appengine', 'set_global_object', 'JsRegExp', 'PyJsException',
    'PyExceptionToJs', 'JsToPyException', 'PyJsSwitchException'
]

# these were defined in base.py
builtins = (
    'true',
    'false',
    'null',
    'undefined',
    'Infinity',
    'NaN',
    'console',
    'String',
    'Number',
    'Boolean',
    'RegExp',
    'Math',
    'Date',
    'Object',
    'Function',
    'Array',
    'Int8Array',
    'Uint8Array',
    'Uint8ClampedArray',
    'Int16Array',
    'Uint16Array',
    'Int32Array',
    'Uint32Array',
    'Float32Array',
    'Float64Array',
    'ArrayBuffer',
    'parseFloat',
    'parseInt',
    'isFinite',
    'isNaN',
    'escape',
    'unescape',
    'encodeURI',
    'decodeURI',
    'encodeURIComponent',
    'decodeURIComponent',
)

#Array, Function, JSON,   Error is done later :)
# also some built in functions like eval...


def set_global_object(obj):
    obj.IS_CHILD_SCOPE = False
    this = This({})
    this.own = obj.own
    this.prototype = obj.prototype
    PyJs.GlobalObject = this
    # make this available
    obj.register('this')
    obj.put('this', this)
    # also add window and set it to be a global object for compatibility
    obj.register('window')
    obj.put('window', this)


scope = dict(zip(builtins, [globals()[e] for e in builtins]))
# Now add errors:
for name, error in ERRORS.items():
    scope[name] = error
#add eval
scope['eval'] = Eval
scope['JSON'] = JSON
JS_BUILTINS = dict((k, v) for k, v in scope.items())



from __future__ import unicode_literals
from pyjsparser.pyjsparserdata import *
from .friendly_nodes import *
import random
import six

if six.PY3:
    from functools import reduce
    xrange = range
    unicode = str
# number of characters above which expression will be split to multiple lines in order to avoid python parser stack overflow
# still experimental so I suggest to set it to 400 in order to avoid common errors
# set it to smaller value only if you have problems with parser stack overflow
LINE_LEN_LIMIT = 400  #  200  # or any other value - the larger the smaller probability of errors :)


class ForController:
    def __init__(self):
        self.inside = [False]
        self.update = ''

    def enter_for(self, update):
        self.inside.append(True)
        self.update = update

    def leave_for(self):
        self.inside.pop()

    def enter_other(self):
        self.inside.append(False)

    def leave_other(self):
        self.inside.pop()

    def is_inside(self):
        return self.inside[-1]


class InlineStack:
    NAME = 'PyJs_%s_%d_'

    def __init__(self):
        self.reps = {}
        self.names = []

    def inject_inlines(self, source):
        for lval in self.names:  # first in first out! Its important by the way
            source = inject_before_lval(source, lval, self.reps[lval])
        return source

    def require(self, typ):
        name = self.NAME % (typ, len(self.names))
        self.names.append(name)
        return name

    def define(self, name, val):
        self.reps[name] = val

    def reset(self):
        self.rel = {}
        self.names = []


class ContextStack:
    def __init__(self):
        self.to_register = set([])
        self.to_define = {}

    def reset(self):
        self.to_register = set([])
        self.to_define = {}

    def register(self, var):
        self.to_register.add(var)

    def define(self, name, code):
        self.to_define[name] = code
        self.register(name)

    def get_code(self):
        code = 'var.registers([%s])\n' % ', '.join(
            repr(e) for e in self.to_register)
        for name, func_code in six.iteritems(self.to_define):
            code += func_code
        return code


def clean_stacks():
    global Context, inline_stack
    Context = ContextStack()
    inline_stack = InlineStack()


def to_key(literal_or_identifier):
    ''' returns string representation of this object'''
    if literal_or_identifier['type'] == 'Identifier':
        return literal_or_identifier['name']
    elif literal_or_identifier['type'] == 'Literal':
        k = literal_or_identifier['value']
        if isinstance(k, float):
            return unicode(float_repr(k))
        elif 'regex' in literal_or_identifier:
            return compose_regex(k)
        elif isinstance(k, bool):
            return 'true' if k else 'false'
        elif k is None:
            return 'null'
        else:
            return unicode(k)

def is_iteration_statement(cand):
    if not isinstance(cand, dict):
        # Multiple statements.
        return False
    return cand.get("type", "?") in {"ForStatement", "ForInStatement", "WhileStatement", "DoWhileStatement"}



def trans(ele, standard=False):
    """Translates esprima syntax tree to python by delegating to appropriate translating node"""
    try:
        node = globals().get(ele['type'])
        if not node:
            raise NotImplementedError('%s is not supported!' % ele['type'])
        if standard:
            node = node.__dict__[
                'standard'] if 'standard' in node.__dict__ else node
        return node(**ele)
    except:
        #print ele
        raise


def limited(func):
    '''Decorator limiting resulting line length in order to avoid python parser stack overflow -
      If expression longer than LINE_LEN_LIMIT characters then it will be moved to upper line
     USE ONLY ON EXPRESSIONS!!! '''

    def f(standard=False, **args):
        insert_pos = len(
            inline_stack.names
        )  # in case line is longer than limit we will have to insert the lval at current position
        # this is because calling func will change inline_stack.
        # we cant use inline_stack.require here because we dont know whether line overflows yet
        res = func(**args)
        if len(res) > LINE_LEN_LIMIT:
            name = inline_stack.require('LONG')
            inline_stack.names.pop()
            inline_stack.names.insert(insert_pos, name)
            res = 'def %s(var=var):\n    return %s\n' % (name, res)
            inline_stack.define(name, res)
            return name + '()'
        else:
            return res

    f.__dict__['standard'] = func
    return f


# ==== IDENTIFIERS AND LITERALS  =======

inf = float('inf')


def Literal(type, value, raw, regex=None):
    if regex:  # regex
        return 'JsRegExp(%s)' % repr(compose_regex(value))
    elif value is None:  # null
        return 'var.get(u"null")'
    # Todo template
    # String, Bool, Float
    return 'Js(%s)' % repr(value) if value != inf else 'Js(float("inf"))'


def Identifier(type, name):
    return 'var.get(%s)' % repr(name)


@limited
def MemberExpression(type, computed, object, property):
    far_left = trans(object)
    if computed:  # obj[prop] type accessor
        # may be literal which is the same in every case so we can save some time on conversion
        if property['type'] == 'Literal':
            prop = repr(to_key(property))
        else:  # worst case
            prop = trans(property)
    else:  # always the same since not computed (obj.prop accessor)
        prop = repr(to_key(property))
    return far_left + '.get(%s)' % prop


def ThisExpression(type):
    return 'var.get(u"this")'


@limited
def CallExpression(type, callee, arguments):
    arguments = [trans(e) for e in arguments]
    if callee['type'] == 'MemberExpression':
        far_left = trans(callee['object'])
        if callee['computed']:  # obj[prop] type accessor
            # may be literal which is the same in every case so we can save some time on conversion
            if callee['property']['type'] == 'Literal':
                prop = repr(to_key(callee['property']))
            else:  # worst case
                prop = trans(
                    callee['property'])  # its not a string literal! so no repr
        else:  # always the same since not computed (obj.prop accessor)
            prop = repr(to_key(callee['property']))
        arguments.insert(0, prop)
        return far_left + '.callprop(%s)' % ', '.join(arguments)
    else:  # standard call
        return trans(callee) + '(%s)' % ', '.join(arguments)


# ========== ARRAYS ============


def ArrayExpression(type, elements):  # todo fix null inside problem
    return 'Js([%s])' % ', '.join(trans(e) if e else 'None' for e in elements)


# ========== OBJECTS =============


def ObjectExpression(type, properties):
    name = None
    elems = []
    after = ''
    for p in properties:
        if p['kind'] == 'init':
            elems.append('%s:%s' % Property(**p))
        else:
            if name is None:
                name = inline_stack.require('Object')
            if p['kind'] == 'set':
                k, setter = Property(
                    **p
                )  # setter is just a lval referring to that function, it will be defined in InlineStack automatically
                after += '%s.define_own_property(%s, {"set":%s, "configurable":True, "enumerable":True})\n' % (
                    name, k, setter)
            elif p['kind'] == 'get':
                k, getter = Property(**p)
                after += '%s.define_own_property(%s, {"get":%s, "configurable":True, "enumerable":True})\n' % (
                    name, k, getter)
            else:
                raise RuntimeError('Unexpected object propery kind')
    definition = 'Js({%s})' % ','.join(elems)
    if name is None:
        return definition
    body = '%s = %s\n' % (name, definition)
    body += after
    body += 'return %s\n' % name
    code = 'def %s():\n%s' % (name, indent(body))
    inline_stack.define(name, code)
    return name + '()'


def Property(type, kind, key, computed, value, method, shorthand):
    if shorthand or computed:
        raise NotImplementedError(
            'Shorthand and Computed properties not implemented!')
    k = to_key(key)
    if k is None:
        raise SyntaxError('Invalid key in dictionary! Or bug in Js2Py')
    v = trans(value)
    return repr(k), v


# ========== EXPRESSIONS ============


@limited
def UnaryExpression(type, operator, argument, prefix):
    a = trans(
        argument, standard=True
    )  # unary involve some complex operations so we cant use line shorteners here
    if operator == 'delete':
        if argument['type'] in ('Identifier', 'MemberExpression'):
            # means that operation is valid
            return js_delete(a)
        return 'PyJsComma(%s, Js(True))' % a  # otherwise not valid, just perform expression and return true.
    elif operator == 'typeof':
        return js_typeof(a)
    return UNARY[operator](a)


@limited
def BinaryExpression(type, operator, left, right):
    a = trans(left)
    b = trans(right)
    # delegate to our friends
    return BINARY[operator](a, b)


@limited
def UpdateExpression(type, operator, argument, prefix):
    a = trans(
        argument, standard=True
    )  # also complex operation involving parsing of the result so no line length reducing here
    return js_postfix(a, operator == '++', not prefix)


@limited
def AssignmentExpression(type, operator, left, right):
    operator = operator[:-1]
    if left['type'] == 'Identifier':
        if operator:
            return 'var.put(%s, %s, %s)' % (repr(to_key(left)), trans(right),
                                            repr(operator))
        else:
            return 'var.put(%s, %s)' % (repr(to_key(left)), trans(right))
    elif left['type'] == 'MemberExpression':
        far_left = trans(left['object'])
        if left['computed']:  # obj[prop] type accessor
            # may be literal which is the same in every case so we can save some time on conversion
            if left['property']['type'] == 'Literal':
                prop = repr(to_key(left['property']))
            else:  # worst case
                prop = trans(
                    left['property'])  # its not a string literal! so no repr
        else:  # always the same since not computed (obj.prop accessor)
            prop = repr(to_key(left['property']))
        if operator:
            return far_left + '.put(%s, %s, %s)' % (prop, trans(right),
                                                    repr(operator))
        else:
            return far_left + '.put(%s, %s)' % (prop, trans(right))
    else:
        raise SyntaxError('Invalid left hand side in assignment!')


six


@limited
def SequenceExpression(type, expressions):
    return reduce(js_comma, (trans(e) for e in expressions))


@limited
def NewExpression(type, callee, arguments):
    return trans(callee) + '.create(%s)' % ', '.join(
        trans(e) for e in arguments)


@limited
def ConditionalExpression(
        type, test, consequent,
        alternate):  # caused plenty of problems in my home-made translator :)
    return '(%s if %s else %s)' % (trans(consequent), trans(test),
                                   trans(alternate))


# ===========  STATEMENTS =============


def BlockStatement(type, body):
    return StatementList(
        body)  # never returns empty string! In the worst case returns pass\n


def ExpressionStatement(type, expression):
    return trans(expression) + '\n'  # end expression space with new line


def BreakStatement(type, label):
    if label:
        return 'raise %s("Breaked")\n' % (get_break_label(label['name']))
    else:
        return 'break\n'


def ContinueStatement(type, label):
    if label:
        return 'raise %s("Continued")\n' % (get_continue_label(label['name']))
    else:
        return 'continue\n'


def ReturnStatement(type, argument):
    return 'return %s\n' % (trans(argument)
                            if argument else "var.get('undefined')")


def EmptyStatement(type):
    return 'pass\n'


def DebuggerStatement(type):
    return 'pass\n'


def DoWhileStatement(type, body, test):
    inside = trans(body) + 'if not %s:\n' % trans(test) + indent('break\n')
    result = 'while 1:\n' + indent(inside)
    return result


def ForStatement(type, init, test, update, body):
    update = indent(trans(update)) if update else ''
    init = trans(init) if init else ''
    if not init.endswith('\n'):
        init += '\n'
    test = trans(test) if test else '1'
    if not update:
        result = '#for JS loop\n%swhile %s:\n%s%s\n' % (
            init, test, indent(trans(body)), update)
    else:
        result = '#for JS loop\n%swhile %s:\n' % (init, test)
        body = 'try:\n%sfinally:\n    %s\n' % (indent(trans(body)), update)
        result += indent(body)
    return result


def ForInStatement(type, left, right, body, each):
    res = 'for PyJsTemp in %s:\n' % trans(right)
    if left['type'] == "VariableDeclaration":
        addon = trans(left)  # make sure variable is registered
        if addon != 'pass\n':
            res = addon + res  # we have to execute this expression :(
        # now extract the name
        try:
            name = left['declarations'][0]['id']['name']
        except:
            raise RuntimeError('Unusual ForIn loop')
    elif left['type'] == 'Identifier':
        name = left['name']
    else:
        raise RuntimeError('Unusual ForIn loop')
    res += indent('var.put(%s, PyJsTemp)\n' % repr(name) + trans(body))
    return res


def IfStatement(type, test, consequent, alternate):
    # NOTE we cannot do elif because function definition inside elif statement would not be possible!
    IF = 'if %s:\n' % trans(test)
    IF += indent(trans(consequent))
    if not alternate:
        return IF
    ELSE = 'else:\n' + indent(trans(alternate))
    return IF + ELSE


def LabeledStatement(type, label, body):
    # todo consider using smarter approach!
    inside = trans(body)
    defs = ''
    if is_iteration_statement(body) and (inside.startswith('while ') or inside.startswith(
            'for ') or inside.startswith('#for')):
        # we have to add contine label as well...
        # 3 or 1 since #for loop type has more lines before real for.
        sep = 1 if not inside.startswith('#for') else 3
        cont_label = get_continue_label(label['name'])
        temp = inside.split('\n')
        injected = 'try:\n' + '\n'.join(temp[sep:])
        injected += 'except %s:\n    pass\n' % cont_label
        inside = '\n'.join(temp[:sep]) + '\n' + indent(injected)
        defs += 'class %s(Exception): pass\n' % cont_label
    break_label = get_break_label(label['name'])
    inside = 'try:\n%sexcept %s:\n    pass\n' % (indent(inside), break_label)
    defs += 'class %s(Exception): pass\n' % break_label
    return defs + inside


def StatementList(lis):
    if lis:  # ensure we don't return empty string because it may ruin indentation!
        code = ''.join(trans(e) for e in lis)
        return code if code else 'pass\n'
    else:
        return 'pass\n'


def PyimportStatement(type, imp):
    lib = imp['name']
    jlib = 'PyImport_%s' % lib
    code = 'import %s as %s\n' % (lib, jlib)
    #check whether valid lib name...
    try:
        compile(code, '', 'exec')
    except:
        raise SyntaxError(
            'Invalid Python module name (%s) in pyimport statement' % lib)
    # var.pyimport will handle module conversion to PyJs object
    code += 'var.pyimport(%s, %s)\n' % (repr(lib), jlib)
    return code


def SwitchStatement(type, discriminant, cases):
    #TODO there will be a problem with continue in a switch statement.... FIX IT
    code = 'while 1:\n' + indent('SWITCHED = False\nCONDITION = (%s)\n')
    code = code % trans(discriminant)
    for case in cases:
        case_code = None
        if case['test']:  # case (x):
            case_code = 'if SWITCHED or PyJsStrictEq(CONDITION, %s):\n' % (
                trans(case['test']))
        else:  # default:
            case_code = 'if True:\n'
        case_code += indent('SWITCHED = True\n')
        case_code += indent(StatementList(case['consequent']))
        # one more indent for whole
        code += indent(case_code)
    # prevent infinite loop and sort out nested switch...
    code += indent('SWITCHED = True\nbreak\n')
    return code


def ThrowStatement(type, argument):
    return 'PyJsTempException = JsToPyException(%s)\nraise PyJsTempException\n' % trans(
        argument)


def TryStatement(type, block, handler, handlers, guardedHandlers, finalizer):
    result = 'try:\n%s' % indent(trans(block))
    # complicated catch statement...
    if handler:
        identifier = handler['param']['name']
        holder = 'PyJsHolder_%s_%d' % (to_hex(identifier),
                                       random.randrange(1e8))
        identifier = repr(identifier)
        result += 'except PyJsException as PyJsTempException:\n'
        # fill in except ( catch ) block and remember to recover holder variable to its previous state
        result += indent(
            TRY_CATCH.replace('HOLDER',
                              holder).replace('NAME', identifier).replace(
                                  'BLOCK', indent(trans(handler['body']))))
    # translate finally statement if present
    if finalizer:
        result += 'finally:\n%s' % indent(trans(finalizer))
    return result


def LexicalDeclaration(type, declarations, kind):
    raise NotImplementedError(
        'let and const not implemented yet but they will be soon! Check github for updates.'
    )


def VariableDeclarator(type, id, init):
    name = id['name']
    # register the name if not already registered
    Context.register(name)
    if init:
        return 'var.put(%s, %s)\n' % (repr(name), trans(init))
    return ''


def VariableDeclaration(type, declarations, kind):
    code = ''.join(trans(d) for d in declarations)
    return code if code else 'pass\n'


def WhileStatement(type, test, body):
    result = 'while %s:\n' % trans(test) + indent(trans(body))
    return result


def WithStatement(type, object, body):
    raise NotImplementedError('With statement not implemented!')


def Program(type, body):
    inline_stack.reset()
    code = ''.join(trans(e) for e in body)
    # here add hoisted elements (register variables and define functions)
    code = Context.get_code() + code
    # replace all inline variables
    code = inline_stack.inject_inlines(code)
    return code


# ======== FUNCTIONS ============


def FunctionDeclaration(type, id, params, defaults, body, generator,
                        expression):
    if generator:
        raise NotImplementedError('Generators not supported')
    if defaults:
        raise NotImplementedError('Defaults not supported')
    if not id:
        return FunctionExpression(type, id, params, defaults, body, generator,
                                  expression) + '\n'
    JsName = id['name']
    PyName = 'PyJsHoisted_%s_' % JsName
    PyName = PyName if is_valid_py_name(PyName) else 'PyJsHoistedNonPyName'
    # this is quite complicated
    global Context
    previous_context = Context
    # change context to the context of this function
    Context = ContextStack()
    # translate body within current context
    code = trans(body)
    # get arg names
    vars = [v['name'] for v in params]
    # args are automaticaly registered variables
    Context.to_register.update(vars)
    # add all hoisted elements inside function
    code = Context.get_code() + code
    # check whether args are valid python names:
    used_vars = []
    for v in vars:
        if is_valid_py_name(v):
            used_vars.append(v)
        else:  # invalid arg in python, for example $, replace with alternatice arg
            used_vars.append('PyJsArg_%s_' % to_hex(v))
    header = '@Js\n'
    header += 'def %s(%sthis, arguments, var=var):\n' % (
        PyName, ', '.join(used_vars) + (', ' if vars else ''))
    # transfer names from Py scope to Js scope
    arg_map = dict(zip(vars, used_vars))
    arg_map.update({'this': 'this', 'arguments': 'arguments'})
    arg_conv = 'var = Scope({%s}, var)\n' % ', '.join(
        repr(k) + ':' + v for k, v in six.iteritems(arg_map))
    # and finally set the name of the function to its real name:
    footer = '%s.func_name = %s\n' % (PyName, repr(JsName))
    footer += 'var.put(%s, %s)\n' % (repr(JsName), PyName)
    whole_code = header + indent(arg_conv + code) + footer
    # restore context
    Context = previous_context
    # define in upper context
    Context.define(JsName, whole_code)
    return 'pass\n'


def FunctionExpression(type, id, params, defaults, body, generator,
                       expression):
    if generator:
        raise NotImplementedError('Generators not supported')
    if defaults:
        raise NotImplementedError('Defaults not supported')
    JsName = id['name'] if id else 'anonymous'
    if not is_valid_py_name(JsName):
        ScriptName = 'InlineNonPyName'
    else:
        ScriptName = JsName
    PyName = inline_stack.require(ScriptName)  # this is unique

    # again quite complicated
    global Context
    previous_context = Context
    # change context to the context of this function
    Context = ContextStack()
    # translate body within current context
    code = trans(body)
    # get arg names
    vars = [v['name'] for v in params]
    # args are automaticaly registered variables
    Context.to_register.update(vars)
    # add all hoisted elements inside function
    code = Context.get_code() + code
    # check whether args are valid python names:
    used_vars = []
    for v in vars:
        if is_valid_py_name(v):
            used_vars.append(v)
        else:  # invalid arg in python, for example $, replace with alternatice arg
            used_vars.append('PyJsArg_%s_' % to_hex(v))
    header = '@Js\n'
    header += 'def %s(%sthis, arguments, var=var):\n' % (
        PyName, ', '.join(used_vars) + (', ' if vars else ''))
    # transfer names from Py scope to Js scope
    arg_map = dict(zip(vars, used_vars))
    arg_map.update({'this': 'this', 'arguments': 'arguments'})
    if id:  # make self available from inside...
        if id['name'] not in arg_map:
            arg_map[id['name']] = PyName
    arg_conv = 'var = Scope({%s}, var)\n' % ', '.join(
        repr(k) + ':' + v for k, v in six.iteritems(arg_map))
    # and finally set the name of the function to its real name:
    footer = '%s._set_name(%s)\n' % (PyName, repr(JsName))
    whole_code = header + indent(arg_conv + code) + footer
    # restore context
    Context = previous_context
    # define in upper context
    inline_stack.define(PyName, whole_code)
    return PyName


LogicalExpression = BinaryExpression
PostfixExpression = UpdateExpression

clean_stacks()

if __name__ == '__main__':
    import codecs
    import time
    import pyjsparser

    c = None  #'''`ijfdij`'''
    if not c:
        with codecs.open("esp.js", "r", "utf-8") as f:
            c = f.read()

    print('Started')
    t = time.time()
    res = trans(pyjsparser.PyJsParser().parse(c))
    dt = time.time() - t + 0.000000001
    print('Translated everyting in', round(dt, 5), 'seconds.')
    print('Thats %d characters per second' % int(len(c) / dt))
    with open('res.py', 'w') as f:
        f.write(res)
